<?php

namespace Drupal\trinion_live_chat\Plugin\rest\resource;

use Bloatless\WebSocket\PushClient;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Database\Connection;
use Drupal\rest\Plugin\ResourceBase;
use Drupal\rest\ResourceResponse;
use Drupal\user\Entity\Role;
use Drupal\user\Entity\User;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\node\Entity\Node;
use Symfony\Component\HttpFoundation\Session\Session;

/**
 * Chat
 *
 * @RestResource (
 *   id = "trinion_live_chat",
 *   label = @Translation("Chat"),
 *   uri_paths = {
 *     "create" = "/rest/chat",
 *   }
 * )
 *
 * @DCG
 * This plugin exposes database records as REST resources. In order to enable it
 * import the resource configuration into active configuration storage. You may
 * find an example of such configuration in the following file:
 * core/modules/rest/config/optional/rest.resource.entity.node.yml.
 * Alternatively you can make use of REST UI module.
 * @see https://www.drupal.org/project/restui
 * For accessing Drupal entities through REST interface use
 * \Drupal\rest\Plugin\rest\resource\EntityResource plugin.
 */
class ChatResource extends ResourceBase {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $dbConnection;

  /**
   * The database connection.
   *
   * @var Session
   */
  protected $session;

  /**
   * Constructs a Drupal\rest\Plugin\rest\resource\EntityResource object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param array $serializer_formats
   *   The available serialization formats.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   * @param \Drupal\Core\Database\Connection $db_connection
   *   The database connection.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, array $serializer_formats, LoggerInterface $logger, Connection $db_connection, Session $session) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $serializer_formats, $logger);
    $this->dbConnection = $db_connection;
    $this->session = $session;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->getParameter('serializer.formats'),
      $container->get('logger.factory')->get('rest'),
      $container->get('database'),
      $container->get('session')
    );
  }

  public function post() {
    $op = \Drupal::request()->get('op');
    switch ($op) {
      case 'send-msg':
        $this->saveMsg();
        $data = [];
        break;
      case 'change-operator-status':
        $data['status'] = $this->changeOperatorStatus();
        break;
      case 'refresh-chats-list':
        $data['chats'] = $this->getChatsList();
        break;
      case 'show-chat':
        $data = $this->getChat();
        break;
      case 'close':
        $data = $this->closeChat();
        break;
      case 'operators-list':
        $data = $this->getOperatorsList();
        break;
      case 'switch-operator':
        $data = $this->switchOperator();
        break;
      default:
        $data = [];
    }
    $response = new ResourceResponse($data);
    $disable_cache = new CacheableMetadata();
    $disable_cache->setCacheMaxAge(0);
    $response->addCacheableDependency($disable_cache);
    return $response;
  }

  /**
   * Отправка сообщения
   */
  public function saveMsg() {
    $user = \Drupal::currentUser();
    $uid = $user->id();
    $sid = $uid ? $uid : $this->session->getId();
    $user = User::load($uid);
    $is_operator = $user->hasPermission('trinion_live_chat operator');
    $msg = json_decode(\Drupal::request()->getContent());
    if (trim($msg->msg) == '')
      return;
    if (isset($msg->chat_id))
      $chat = Node::load($msg->chat_id);
    else {
      $chat = $this->getChatByUser($sid);
    }
    if (!$chat)
      $chat = $this->createChat($uid, $sid);
    $chat_id = $chat->id();
    $operator = $this->getChatOperator($chat);
    $operator_uid = $operator['operator'];
    $operator = User::load($operator_uid);
    // o - operator
    // u - user
    $this->addMessage($chat, $msg->msg, $is_operator ? 'o' : 'u', $operator_uid);
    $time = date('H:i j.m.y', time());

    $chat_user_uid = $chat->get('uid')->getString();
    $chat_user_name = 'Пользователь';
    if ($chat_user_uid && is_numeric($chat_user_uid)) {
      $chat_user = User::load($chat_user_uid);
      $chat_user_name = $chat_user->get('field_tb_nick_name')->getString();
    }

    $pushClient = new PushClient(DRUPAL_ROOT . '/' . \Drupal::service('module_handler')->getModule('trinion_base')->getPath() . '/wserver/phpwss.sock');
    $operator_name = $operator ? $this->operatorName($operator) : 'Оператор';

    //send to operator
    $pushClient->sendToApplication('ws', [
      'action' => 'newChatMessage',
      'data' => [
        'sid' => $operator_uid,
        'uid' => $operator_uid,
        'msg' => nl2br($msg->msg),
        'who' => 'o',
        'type' => $uid == $operator_uid ? 'out' : 'in',
        'time' => $time,
        'chat_id' => $chat_id,
        'operator' => $operator_name,
        'user_name' => $uid == $operator_uid ? $operator_name : $chat_user_name,
      ],
    ]);

    //send to user
    $pushClient->sendToApplication('ws', [
      'action' => 'newChatMessage',
      'data' => [
        'sid' => $chat->get('field_tlc_sid')->getString(),
        'uid' => $chat->get('uid')->getString(),
        'msg' => nl2br($msg->msg),
        'who' => 'u',
        'type' => $uid == $operator_uid ? 'in' : 'out',
        'time' => $time,
        'chat_id' => $chat_id,
        'operator' => $operator_name,
        'user_name' => $uid == $operator_uid ? $operator_name : $chat_user_name,
      ],
    ]);

    // $pushClient->sendToApplication('ws', [
    //   'action' => 'newChatMessage',
    //   'data' => [
    //     'sid' => $sid,
    //     'uid' => $uid,
    //     'msg' => nl2br($msg->msg),
    //     'who' => 'u',
    //     'type' => 'out',
    //     'time' => $time,
    //     'chat_id' => $chat_id,
    //     'operator' => $operator ? $this->operatorName($operator) : 'Оператор',
    //     'user_name' => $chat_user_name,
    //   ],
    // ]);

    // if ($is_operator) {
    //   $pushClient->sendToApplication('ws', [
    //     'action' => 'newChatMessage',
    //     'data' => [
    //       'sid' => $chat->get('field_tlc_sid')->getString(),
    //       'uid' => $chat->get('uid')->getString(),
    //       'msg' => nl2br($msg->msg),
    //       'who' => 'o',
    //       'type' => 'in',
    //       'time' => $time,
    //       'chat_id' => $chat_id,
    //       'operator' => $operator ? $this->operatorName($operator) : 'Оператор',
    //       'user_name' => $chat_user_name,
    //     ],
    //   ]);
    // }
    // else {
    //   $pushClient->sendToApplication('ws', [
    //     'action' => 'newChatMessage',
    //     'data' => [
    //       'sid' => $operator_uid,
    //       'uid' => $operator_uid,
    //       'msg' => nl2br($msg->msg),
    //       'who' => 'o',
    //       'type' => 'in',
    //       'time' => $time,
    //       'chat_id' => $chat_id,
    //       'operator' => $operator ? $this->operatorName($operator) : 'Оператор',
    //       'user_name' => $chat_user_name,
    //     ],
    //   ]);
    // }
  }

  /**
   * Изменение статуса оператора
   */
  public function changeOperatorStatus() {
    $user = \Drupal::currentUser();
    $uid = $user->id();
    $user = User::load($uid);
    $is_operator = $user->hasPermission('trinion_live_chat operator');
    if ($is_operator) {
      $data = json_decode(\Drupal::request()->getContent());
      $user->field_tlc_operator_zanyat = !$data->status;
      $user->save();
    }
    return $user->get('field_tlc_operator_zanyat')->getString();
  }

  /**
   * Получение чатов, где оператор участвует
   */
  public function getChatsList() {
    $user = \Drupal::currentUser();
    $uid = $user->id();
    $query = \Drupal::entityQuery('node')
      ->accessCheck(TRUE)
      ->condition('type', 'trinion_chat')
      ->condition('field_tlc_otvetstvennyy', $uid);
    $res = $query->accessCheck()->execute();
    $chats = [];
    foreach ($res as $chat_nid) {
      $chat = Node::load($chat_nid);
      $msg_cnt = count($chat->get('field_tlc_soobsheniya'));
      if ($msg_cnt == 0)
        continue;
      $chat_user_uid = $chat->get('uid')->getString();
      $chat_user_name = 'Аноним';
      if ($chat_user_uid && is_numeric($chat_user_uid)) {
        $chat_user = User::load($chat_user_uid);
        $chat_user_name = $chat_user->get('field_tb_nick_name')->getString();
      }
      $last_msg = json_decode($chat->get('field_tlc_soobsheniya')[$msg_cnt - 1]->getString());
      $chats[] = [
        'time' => $last_msg->time,
        'user_name' => $chat_user_name,
        'msg' => $last_msg->msg,
        'chat_nid' => $chat_nid,
        'display_time' => date('H:i:s j.m.y', $last_msg->time),
      ];
    }
    array_multisort($chats);
    return $chats;
  }

  /**
   * Закрытие чата
   */
  public function closeChat() {
    $data = json_decode(\Drupal::request()->getContent());
    if (!empty($data->id)) {
      if ($uid = \Drupal::currentUser()->id()) {
        $user = User::load($uid);
        if ($user->hasPermission('trinion_live_chat operator')) {
          $close_msg = 'Оператор закрыл чат';
          $chat_id = $data->id;
          $chat = Node::load($chat_id);
          $this->addMessage($chat, $close_msg, 'o', $uid);
          $chat->field_tlc_otvetstvennyy = NULL;
          $chat->save();

          $time = date('H:i j.m.y', time());
          $pushClient = new PushClient(DRUPAL_ROOT . '/' . \Drupal::service('module_handler')->getModule('trinion_base')->getPath() . '/wserver/phpwss.sock');

          $pushClient->sendToApplication('ws', [
            'action' => 'newChatMessage',
            'data' => [
              'sid' => $uid,
              'uid' => $uid,
              'msg' => nl2br($close_msg),
              'who' => 'u',
              'type' => 'out',
              'time' => $time,
              'chat_id' => $chat_id,
            ],
          ]);
        }
      }
    }
    return [];
  }

  /**
   * Получение чата
   */
  public function getChat() {
    if (!$this->session->isStarted()) {
      $this->session->set('session_started', TRUE);
      $this->session->save();
    }

    $user = \Drupal::currentUser();
    $uid = $user->id();
    $user = User::load($uid);
    $is_operator = $user->hasPermission('trinion_live_chat operator');

    $data = json_decode(\Drupal::request()->getContent());
    if (!empty($data->id)) {
      $chat = Node::load($data->id);
    }
    else {
      $sid = $uid ? $uid : $this->session->getId();
      $chat = $this->getChatByUser($sid);
    }

    $msgs = [];
    $chat_user_name = 'Аноним';
    $chat_user_uid = $chat->get('uid')->getString();
      $client_id = 0;
      if ($chat_user_uid) {
        $chat_user = User::load($chat_user_uid);
        $chat_user_name = $chat_user->get('field_tb_nick_name')->getString();
        $client_id = $chat_user->id();

        if (!$chat_user_name)
          $chat_user_name = $chat_user->label();
      }
    if ($chat) {
      foreach ($chat->get('field_tlc_soobsheniya') as $msg) {
        $msg = json_decode($msg->getString(), 1);
        $msg['time'] = date('H:i j.m.y', $msg['time']);
        if ($is_operator)
          $msg['type'] = $msg['who'] == 'o' ? 'out' : 'in';
        else
          $msg['type'] = $msg['who'] == 'o' ? 'in' : 'out';
        $msg['chat_id'] = $chat->id();
        if ($msg['oper_uid'] && $msg_operator = User::load($msg['oper_uid']))
          $msg['user_name'] = $this->operatorName($msg_operator);
        else
          $msg['user_name'] = $chat_user_name;
        $msgs[] = $msg;
      }
      
      $operator_id = $chat->get('field_tlc_otvetstvennyy')->getString();
      $operator = User::load($operator_id);
    }
    else
      $operator_id = 0;

    return [
      'msgs' => $msgs,
      'user_name' => $chat_user_name,
      'operator' => empty($operator) ? 'Оператор' : $this->operatorName($operator),
      'operator_id' => $operator_id,
      'client_id' => $client_id,
    ];
  }

  public function getChatByUser($sid) {
    $query = \Drupal::entityQuery('node')
      ->accessCheck(TRUE)
      ->condition('type', 'trinion_chat')
      ->condition('field_tlc_sid', $sid);
    $res = $query->execute();
    return $res ? Node::load(reset($res)) : FALSE;
  }

  public function createChat($uid, $sid) {
    $node = Node::create([
      'type' => 'trinion_chat',
      'title' => $this->getNextChatNumber(),
      'uid' => $uid,
      'field_tlc_sid' => $sid,
      'status' => 1,
    ]);
    $node->save();
    return $node;
  }

  public function getNextChatNumber() {
    $start_nomer = 1;
    $query = \Drupal::database()->select('node_field_data', 'n')
      ->condition('n.type', 'trinion_chat');
    $query->addField('n', 'title');
    $query->addExpression('CAST (n.title AS UNSIGNED)', 't');
    $query->orderBy('t', 'DESC');
    $query->range(0, 1);
    $res = $query->execute()->fetchField();
    return $res < $start_nomer ? $start_nomer : $res + 1;
  }

  public function addMessage(Node $chat, $msg, $who, $operator) {
    $message['time'] = time();
    $message['who'] = $who;
    $message['msg'] = $msg;
    if ($who == 'o')
      $message['oper_uid'] = $operator;
    $chat->field_tlc_soobsheniya[] = json_encode($message);
    $chat->field_tlc_otvetstvennyy = $operator;
    $chat->save();
  }

  public function getChatOperator(Node $chat) {
    $operator = $chat->get('field_tlc_otvetstvennyy')->getString();
    if (empty($operator)) {
      $operators = $this->getOperatorsList();
      $uids = array_keys($operators);
      $query = \Drupal::database()->select('users_field_data', 'o')
        ->condition('o.uid', $uids, 'IN');
      $query->leftJoin('node__field_tlc_otvetstvennyy', 'no', 'no.field_tlc_otvetstvennyy_target_id = o.uid');
      $query->addField('o', 'uid');
      $query->addExpression('SUM(no.bundle IS NOT NULL)', 'cnt');
      $query->leftJoin('user__field_tlc_operator_zanyat', 'ob', 'ob.entity_id = o.uid');
      $query->condition('field_tlc_operator_zanyat_value', 0);
      $query->groupBy('o.uid');
      $query->orderBy('cnt');
      $query->range(0, 1);
      $operator = $query->execute()->fetchField();
      $new = 1;
    }
    else
      $new = 0;
    return [
      'operator' => $operator,
      'new' => $new,
    ];
  }

  /**
   * Список рабочих операторов
   */
  public function getOperatorsList() {
    $query = \Drupal::entityQuery('user_role');
    $res = $query->accessCheck()->execute();
    $roles = [];
    foreach (Role::loadMultiple($res) as $role) {
      if ($role->hasPermission('trinion_live_chat operator'))
        $roles[] = $role->id();
    }
    $query = \Drupal::entityQuery('user')
      ->condition('roles', $roles, 'IN');
    $group = $query->orConditionGroup();
    $group->condition('field_tlc_operator_zanyat', 0);
    $group->condition('field_tlc_operator_zanyat', NULL, 'IS NULL');
    $query->condition($group);
    $res = $query->accessCheck()->execute();
    $data = [];
    foreach (User::loadMultiple($res) as $user) {
      $data[$user->id()] = $user->get('field_tb_nick_name')->getString();
    }
    return $data;
  }

  /**
   * Смена оператора
   */
  public function switchOperator() {
    $data = json_decode(\Drupal::request()->getContent());
    if (isset($data->chat_id)) {
      $chat = Node::load($data->chat_id);
      $chat->field_tlc_otvetstvennyy = $data->uid;
      $chat->save();
      $pushClient = new PushClient(DRUPAL_ROOT . '/' . \Drupal::service('module_handler')->getModule('trinion_base')->getPath() . '/wserver/phpwss.sock');

      $time = date('H:i j.m.y', time());
      $pushClient->sendToApplication('ws', [
        'action' => 'newChatMessage',
        'data' => [
          'sid' => $data->uid,
          'uid' => $data->uid,
          'msg' => 'Вам передали чат',
          'who' => 'o',
          'type' => 'out',
          'time' => $time,
          'chat_id' => $data->chat_id,
        ],
      ]);
    }
  }

  /**
   * формирую имя оператора для показа в чате
   * @param $entity
   * @return string|null
   */
  public function operatorName($entity) {
    $name = $entity->get('field_tb_nick_name')->getString();
    if ($name == '')
      $name = $entity->label();
    return $name;
  }

}
